import os

PROJECT_ROOT = os.path.abspath(os.path.dirname(__file__))
CORPUS_FOLDER = os.path.join(PROJECT_ROOT, "..", "corpus")

PATH_SEPARATOR = "::"

IMPORT_MATPLOTLIB = False

DEBUG = True

LOGGING_CONFIG = {
    # TODO
}


