from enum import Enum


class MatrixKeys(Enum):
    PITCH = 0
    VELOCITY = 1
    CHANNEL = 2
    REL_ONSET = 3
    ABS_ONSET = 4
    REL_DURATION = 5
    ABS_DURATION = 6
    TEMPO = 7
    BAR_NUMBER = 8
    TRACK_NAME = 9
